%%% $Id: rd_common_resource_manager.erl,v 1.2 2004/08/04 16:55:47 mlogan Exp $
%%%-------------------------------------------------------------------
%%% File    : rd_common_resource_manager.erl
%%% Author  : Martin J. Logan <martin@mecha.erlware.com>
%%%
%%% @doc  Monitors common_instance()s. See resource_discovery.erl docs for details on a common_instance() definition.
%%% @end
%%%
%%% Created : 30 Jul 2004 by Martin J. Logan <martin@mecha.erlware.com>
%%%-------------------------------------------------------------------
-module(rd_common_resource_manager).

-behaviour(gen_server).
%%--------------------------------------------------------------------
%% Include files
%%--------------------------------------------------------------------

%%--------------------------------------------------------------------
%% External exports
%%--------------------------------------------------------------------
-export([
         start_link/0,
         stop/0,
         add_targettypes/1,
         remove_targettype/1,
         resource_up/2
         ]).

%%--------------------------------------------------------------------
%% gen_server callbacks
%%--------------------------------------------------------------------
-export([init/1, handle_call/3, handle_cast/2, handle_info/2, terminate/2, code_change/3]).

%%--------------------------------------------------------------------
%% record definitions
%%--------------------------------------------------------------------
-record(state, {target_types, monitored_resources}).
%%--------------------------------------------------------------------
%% macro definitions
%%--------------------------------------------------------------------
-define(SERVER, ?MODULE).

%%====================================================================
%% External functions
%%====================================================================
%%--------------------------------------------------------------------
%% @doc Starts the server.
%% @spec start_link() -> {ok, pid()} | {error, Reason}
%% @end
%%--------------------------------------------------------------------
start_link() ->
    gen_server:start_link({local, ?SERVER}, ?MODULE, [], []).

%%--------------------------------------------------------------------
%% @doc Stops the server.
%% @spec stop() -> ok
%% @end
%%--------------------------------------------------------------------
stop() ->
    gen_server:cast(?SERVER, stop).

%%------------------------------------------------------------------------------
%% @doc Adds to the list of target types.
%% <pre>
%%
%% Expects: 
%%  TargetTypes  - A type or list of type()s to be discovered in the node cluster
%%
%% Types:
%%  TargetTypes = [type()] | type()
%%  Reason = term()
%%
%% </pre>
%%
%% @spec add_targettypes(TargetTypes) -> ok | {error, Reason}
%% @end
%%------------------------------------------------------------------------------
add_targettypes(TargetTypes) when is_list(TargetTypes) -> 
    gen_server:call(?SERVER, {add_targettypes, TargetTypes});
add_targettypes(TargetType) -> 
    add_targettypes([TargetType]).
    
%%------------------------------------------------------------------------------
%% @doc Gets the number of resource types.
%% @spec remove_targettype(type()) -> {ok, integer()}
%% @end
%%------------------------------------------------------------------------------
remove_targettype(Type) ->
    gen_server:cast(?SERVER, {remove_targettype, Type}).

%%--------------------------------------------------------------------
%% @doc called by resource discovery when a new resource is found on the network.
%% @spec resource_up(type(), common_instance()) -> ok
%% @end
%%--------------------------------------------------------------------
resource_up(Type, CommonInstance) ->
    error_logger:info_msg("~p:resource_up Type ~p Instance ~p~n", [?MODULE, Type, CommonInstance]),
    gen_server:cast(?SERVER, {resource_up, {Type, CommonInstance}}).

%%====================================================================
%% Server functions
%%====================================================================

%%--------------------------------------------------------------------
%% Function: init/1
%% Description: Initiates the server
%% Returns: {ok, State}          |
%%          {ok, State, Timeout} |
%%          ignore               |
%%          {stop, Reason}
%%--------------------------------------------------------------------
init([]) ->
    error_logger:info_msg("~p:init~n", [?MODULE]),
    resource_discovery:add_callback(?MODULE),
    {ok, #state{target_types = [], monitored_resources = dict:new()}}.

%%--------------------------------------------------------------------
%% Function: handle_call/3
%% Description: Handling call messages
%% Returns: {reply, Reply, State}          |
%%          {reply, Reply, State, Timeout} |
%%          {noreply, State}               |
%%          {noreply, State, Timeout}      |
%%          {stop, Reason, Reply, State}   | (terminate/2 is called)
%%          {stop, Reason, State}            (terminate/2 is called)
%%--------------------------------------------------------------------
handle_call({add_targettypes, Types}, From, #state{target_types = TT} = State) ->
    resource_discovery:add_targettypes(Types),
    {reply, ok, State#state{target_types = sets:to_list(sets:from_list((TT ++ Types)))}}.

%%--------------------------------------------------------------------
%% Function: handle_cast/2
%% Description: Handling cast messages
%% Returns: {noreply, State}          |
%%          {noreply, State, Timeout} |
%%          {stop, Reason, State}            (terminate/2 is called)
%%--------------------------------------------------------------------
handle_cast(stop, State) ->
    {stop, normal, State};
handle_cast({remove_targettype, Type}, #state{target_types = TT} = State) ->
    {noreply, State#state{target_types = lists:delete(Type, TT)}};
handle_cast({resource_up, {Type, Instance} = Resource}, State) when Instance == node() ->
    {noreply, State};
handle_cast({resource_up, {Type, Instance} = Resource}, #state{target_types = TT, monitored_resources = MRDict} = State) ->
    case lists:member(Type, TT) of
        true  -> 
            case dict:find(Instance, MRDict) of
                {ok, Types} -> 
                    NewMRDict = case lists:member(Type, Types) of
                                    true  -> MRDict;
                                    false -> dict:append(Instance, Type, MRDict)
                                end,
                    {noreply, State#state{monitored_resources = NewMRDict}};
                error ->
                    NewMRDict = dict:append(Instance, Type, MRDict),
                    monitor(Instance),
                    {noreply, State#state{monitored_resources = NewMRDict}}
            end;
        false -> 
            {noreply, State}
    end.

%%--------------------------------------------------------------------
%% Function: handle_info/2
%% Description: Handling all non call/cast messages
%% Returns: {noreply, State}          |
%%          {noreply, State, Timeout} |
%%          {stop, Reason, State}            (terminate/2 is called)
%%--------------------------------------------------------------------
handle_info({nodedown, Node}, #state{monitored_resources = MRDict} = State) ->
    Types = dict:fetch(Node, MRDict),
    error_logger:info_msg("rd_common_resource_manager:handle_info/2 Node CommonResource Down Types ~p Instance ~p~n", [Types, Node]),
    lists:foreach(fun(Type) -> resource_discovery:remove_resource(Type, Node) end, Types),
    {noreply, State#state{monitored_resources = dict:erase(Node, MRDict)}};
handle_info({'DOWN', MonitorReference, process, DeadProcess, Info} = Msg, #state{monitored_resources = MRDict} = State) ->
    Types = dict:fetch(DeadProcess, MRDict),
    error_logger:info_msg("rd_common_resource_manager:handle_info/2 Node CommonResource Down Types ~p Instance ~p~n", [Types, DeadProcess]),
    lists:foreach(fun(Type) -> resource_discovery:remove_resource(Type, DeadProcess) end, Types),
    {noreply, State#state{monitored_resources = dict:erase(DeadProcess, MRDict)}}.
    

%%--------------------------------------------------------------------
%% Function: terminate/2
%% Description: Shutdown the server
%% Returns: any (ignored by gen_server)
%%--------------------------------------------------------------------
terminate(Reason, State) ->
    ok.

%%--------------------------------------------------------------------
%% Func: code_change/3
%% Purpose: Convert process state when code is changed
%% Returns: {ok, NewState}
%%--------------------------------------------------------------------
code_change(OldVsn, State, Extra) ->
    {ok, State}.

%%====================================================================
%%% Internal functions
%%====================================================================

monitor(CommonInstance) when is_pid(CommonInstance) -> erlang:monitor(process, CommonInstance);
monitor(CommonInstance)                             -> erlang:monitor_node(CommonInstance, true).
