%%%----------------------------------------------------------------------
%%% File    : rd_core.erl
%%% Author  : Martin J. Logan <martin@gdubya.botomayo>
%%% @doc 
%%% <p>The running process that is responsable for managing the store and 
%%% processing user requests. Stores data in a dict within loop data 
%%% for speed.</p>
%%% @end
%%%----------------------------------------------------------------------

-module(rd_core).

-behaviour(gen_server).

%% Gen_server callbacks
-export([
	 init/1, 
         handle_call/3, 
         handle_cast/2, 
         handle_info/2, 
         code_change/3, 
         terminate/2
	]).

% Macros
-define(SERVER, ?MODULE).
-define(RD, resource_discovery).

% Records
-record(state, {resources, contact_node, callback}).


%%%----------------------------------------------------------------------
%%% Callback functions from gen_server
%%%----------------------------------------------------------------------

%%----------------------------------------------------------------------
%% Func: init/1
%% Returns: {ok, State}          |
%%          {ok, State, Timeout} |
%%          ignore               |
%%          {stop, Reason}
%%----------------------------------------------------------------------
init(_) -> 
    timer:sleep(500),
    error_logger:info_msg("rd_core:init~n"),
    ConfigModule = 
        case application:get_env(config_module) of
            {ok, Module} -> Module;
            undefined    -> application
        end,

    Callback = 
        case ConfigModule:get_env(resource_discovery, callback) of
	        {ok, CB} when is_list(CB) -> CB;
	        {ok, CB}                  -> [CB];
	        undefined                 -> []
        end,
 
    ContactNode = case lists:keysearch(contact_node, 1, init:get_arguments()) of
                      {value, {contact_node, [I_ContactNode]}} -> {ok, list_to_atom(I_ContactNode)};
                      _                                        -> ConfigModule:get_env(contact_node)
                  end,

    case ContactNode of
        {ok, Node} -> 
            error_logger:info_msg("rd_core:init Contact node to ping is ~p~n", [Node]),
            % If a contact node is specified then it must be reachable.
            pong = 
                case net_adm:ping(Node) of
	                pang -> error_logger:error_msg("rd_core:init bad contact node. pang~n"), pang;
	                pong -> pong
                end,
            {ok, #state{resources = rd_store:new_RL(), contact_node = Node, callback = Callback}};
        undefined  -> 
            {ok, #state{resources = rd_store:new_RL(), contact_node = null, callback = Callback}}
    end.

%%----------------------------------------------------------------------
%% Func: handle_call/3
%% Returns: {reply, Reply, State}          |
%%          {reply, Reply, State, Timeout} |
%%          {noreply, State}               |
%%          {noreply, State, Timeout}      |
%%          {stop, Reason, Reply, State}   | (terminate/2 is called)
%%          {stop, Reason, State}            (terminate/2 is called)
%%----------------------------------------------------------------------

%% Get a resource
handle_call({get_resource, Type}, From, State) ->
    case rd_store:round_robin_lookup(Type, State#state.resources) of 
	{ok, {Resource, NewResources}} -> {reply, {ok, Resource}, State#state{resources = NewResources}};
	{error, _}                     -> {reply, {error, no_resources}, State}
    end;

%% Get a resource based on an index
handle_call({get_resource, Type, Index}, From, State) ->
    case rd_store:index_lookup(Type, Index, State#state.resources) of 
	{ok, Resource} -> {reply, {ok, Resource}, State};
	{error, _}     -> {reply, {error, no_resources}, State}
    end;

handle_call({get_all_resources, Type}, From, State) ->
    {reply, {ok, rd_store:lookup_all_resources(Type, State#state.resources)}, State};

%% Add a new list of localtypes target types and inform the network.
handle_call({full_add, RawLocalTypes, RawTargetTypes}, From, State) ->
    {ok, LocalTypes}  = format_localtypes(RawLocalTypes),
    {ok, TargetTypes} = format_targettypes(RawTargetTypes),
    rd_store:store_localtypes(LocalTypes),
    rd_store:store_targettypes(TargetTypes),
    inform_nodes(),
    {reply, ok, State};

%% Add a new list of target types
handle_call({add_targettypes, RawTargetTypes}, From, State) ->
    {ok, TargetTypes} = format_targettypes(RawTargetTypes),
    rd_store:store_targettypes(TargetTypes),
    {reply, ok, State};

%% Add local types
handle_call({add_localtypes, RawLocalTypes}, From, State) ->
    {ok, LocalTypes} = format_localtypes(RawLocalTypes),
    rd_store:store_localtypes(LocalTypes),
    {reply, ok, State};

%% Get number of resources.
handle_call({get_num_resource, Type}, From, State) ->
    {reply, {ok, rd_store:lookup_num_resources(Type, State#state.resources)}, State};

%% Remove a local type from the local store.
handle_call({remove_localtype, LocalType}, From, State) ->
    error_logger:info_msg("remove_localtype: ~p~n", [LocalType]),
    rd_store:delete_localtype(LocalType),
    {reply, ok, State};

%% Remove a target type from the local store.
handle_call({remove_targettype, TargetType}, From, State) ->
    error_logger:info_msg("remove_targettype: ~p~n", [TargetType]),
    rd_store:delete_targettype(TargetType),
    {reply, ok, State};

%% Get the number of types on the network 
handle_call(get_num_types, From, State) ->
    {reply, {ok, rd_store:lookup_num_types(State#state.resources)}, State};

%% Get a list of types that we have.
handle_call(get_types, From, State) ->
    {reply, {ok, rd_store:lookup_types(State#state.resources)}, State};

%% Prompt a sending of all local and target types.
%% Ensures a totaly up to date resource list by starting from scratch.
handle_call(sync_inform_network, From, State) ->
    NewResources = sync_inform_nodes(rd_store:new_RL(), State#state.callback),
    {reply, ok, State#state{resources = NewResources}};


%% Prompt a sending of all local and target types.
handle_call(inform_network, From, State) ->
    {reply, inform_nodes(), State};

%% Ping the node contact if there is one
handle_call(contact_node, From, State) when State#state.contact_node =:= null ->
    {reply, {null, pang}, State};
handle_call(contact_node, From, State) ->
    Node = State#state.contact_node,
    {reply, {Node, net_adm:ping(Node)}, State};

%% Add a callback function to resource discovery.
handle_call({add_callback, Module}, From, State) ->
    NewState = State#state{callback = [Module|lists:delete(Module, State#state.callback)]},
    {reply, ok, NewState};

handle_call({call_resource_up, RemoteResources, RemoteTargetTypes}, From, State) ->
    {RemotePid, _} = From,
    error_logger:info_msg("rd_core:call_resource_up: from ~p ~p~n", [node(RemotePid), RemoteResources]),
    % Does the local node posess the remote nodes target types?
    CanServe = 
        fun({LocalType, LocalResource} = FullResource, Acc) ->
                case can_serve(LocalType, RemoteTargetTypes) of
                    true  -> [FullResource|Acc];
                    false -> Acc
                end
        end,
    CanServeList = lists:foldl(CanServe, [], rd_store:lookup_localtypes()), 
    gen_server:reply(From, {ok, CanServeList}),
    
    % Does the remote node have any resources that we care about?
    CanUse = 
        fun({RemoteType, RemoteResource}, Resources) -> 
                case can_use(rd_store:lookup_targettypes(), RemoteType) of
                    true  -> 
                        do_callback(State#state.callback, RemoteType, RemoteResource),
                        rd_store:store_resource(RemoteType, RemoteResource, Resources);
                    false -> 
                        Resources
                end
        end,
    {noreply, State#state{resources = lists:foldl(CanUse, State#state.resources, RemoteResources)}}.


%%----------------------------------------------------------------------
%% Func: handle_cast/2
%% Returns: {noreply, State}          |
%%          {noreply, State, Timeout} |
%%          {stop, Reason, State}            (terminate/2 is called)
%%----------------------------------------------------------------------

% Remove a resource from the local app pool.
handle_cast({remove_resource, Type, Node}, State) ->
    error_logger:info_msg("remove_resource: ~p ~p ~n", [Type, Node]),
    Resources = rd_store:delete_resource(Type, Node, State#state.resources),
    {noreply, State#state{resources = Resources}};

% Add a remote resource to this managers pool of resources..
handle_cast({add_resource_on, RemoteType, RemoteResource}, State) ->
    error_logger:info_msg("rd_core:add_resource_on: ~p ~p~n", [RemoteType, RemoteResource]),
    NewResourceList = rd_store:store_resource(RemoteType, RemoteResource, State#state.resources),
    do_callback(State#state.callback, RemoteType, RemoteResource),
    {noreply, State#state{resources = NewResourceList}}; 

% A resource is now up do they care about us?
handle_cast({cast_resource_up, From, RemoteResources, RemoteTargetTypes}, State) ->
    error_logger:info_msg("rd_core:resource_up: ~p ~p~n", [From, RemoteResources]),

    % Are we one of the remote nodes target types?
    CanServe = 
        fun({LocalType, LocalResource} = FullResource) ->
                case can_serve(LocalType, RemoteTargetTypes) of
                    true  -> add_resource_on(From, FullResource);
                    false -> false
                end
        end,
    lists:foreach(CanServe, rd_store:lookup_localtypes()), 

    % Does the remote node have any resources that we care about?
    CanUse = 
        fun({RemoteType, RemoteResource}, Resources) -> 
                case can_use(rd_store:lookup_targettypes(), RemoteType) of
                    true  -> 
                        do_callback(State#state.callback, RemoteType, RemoteResource),
                        rd_store:store_resource(RemoteType, RemoteResource, Resources);
                    false -> 
                        Resources
                end
        end,
    {noreply, State#state{resources = lists:foldl(CanUse, State#state.resources, RemoteResources)}};

% Stop
handle_cast({stop, Reason}, State) ->
    {stop, Reason, State}.


%%----------------------------------------------------------------------
%% Func: handle_info/2
%% Returns: {noreply, State}          |
%%          {noreply, State, Timeout} |
%%          {stop, Reason, State}            (terminate/2 is called)
%%----------------------------------------------------------------------
handle_info(Info, State) ->
    {noreply, State}.

%%----------------------------------------------------------------------
%% Func: terminate/2
%% Purpose: Shutdown the server
%% Returns: any (ignored by gen_server)
%%----------------------------------------------------------------------
terminate(Reason, State) ->
    ok.

%%--------------------------------------------------------------------
%% Func: code_change/3
%% Purpose: Convert process state when code is changed
%% Returns: {ok, NewState}
%%--------------------------------------------------------------------
code_change(OldVsn, State, Extra) ->
    {ok, State}.

%%%----------------------------------------------------------------------
%%% Internal functions
%%%----------------------------------------------------------------------

%%----------------------------------------------------------------------
%% Notifies the requesting entity that we are on of the targeted types and provides
%% our node() and type.
%% Expects: ClientNode - The clients node identifier.		      
%%          LocalType  - This apps' type.
%%----------------------------------------------------------------------
add_resource_on(RemoteRD, {LocalType, LocalResource}) ->
    gen_server:cast({?RD, RemoteRD}, {add_resource_on, LocalType, LocalResource}).

%%----------------------------------------------------------------------
%% Asyncronously inform another node as to this nodes local and target types.
%% call_resource_up(Node, LocalTypes, TargetTypes) -> RemoteResourceList
%%----------------------------------------------------------------------
call_resource_up(RemoteNode, LocalTypes, TargetTypes) ->
    case catch gen_server:call({?RD, RemoteNode}, {call_resource_up, LocalTypes, TargetTypes}) of
        {ok, RemoteResourceList} -> RemoteResourceList;
        _Other                   -> []
    end.

%%----------------------------------------------------------------------
%% Asyncronously inform another node as to this nodes local and target types.
%% cast_resource_up(Node, LocalTypes, TargetTypes) -> ok 
%%----------------------------------------------------------------------
cast_resource_up(RemoteNode, LocalTypes, TargetTypes) ->
    gen_server:cast({?RD, RemoteNode}, {cast_resource_up, node(), LocalTypes, TargetTypes}).
    
%% This determines if the callback function is to be called
%% Returns: void()
do_callback([Callback|T], RemoteType, RemoteResource) ->
    catch Callback:resource_up(RemoteType, RemoteResource),
    do_callback(T, RemoteType, RemoteResource);
do_callback([], RemoteType, RemoteResource) ->
    null.

%% This informs all other nodes as to your existance.
%% Returns: ok
inform_nodes() ->
    LocalTypes  = rd_store:lookup_localtypes(),
    TargetTypes = rd_store:lookup_targettypes(),
    lists:foreach(fun(RemoteNode) -> cast_resource_up(RemoteNode, LocalTypes, TargetTypes) end, [node()|nodes()]).

%% This informs all other nodes as to your existance syncronously.
%% Returns: RemoteResourceList
sync_inform_nodes(Resources, CallBack) ->
    LocalTypes  = rd_store:lookup_localtypes(),
    TargetTypes = rd_store:lookup_targettypes(),
    % Make sure that local resources are accounted for
    CanUse = 
        fun({LocalType, LocalResource}, I_Resources) -> 
                case can_use(TargetTypes, LocalType) of
                    true  -> 
                        do_callback(CallBack, LocalType, LocalResource),
                        rd_store:store_resource(LocalType, LocalResource, I_Resources);
                    false -> 
                        I_Resources
                end
        end,
    NewResources = lists:foldl(CanUse, Resources, LocalTypes),
    Nodes = nodes(),
    error_logger:info_msg("calling resource up on ~p~n", [Nodes]),
    sync_inform_nodes(NewResources, CallBack, Nodes, LocalTypes, TargetTypes).

% Inform nodes one by one.
sync_inform_nodes(Resources, CallBack, [Node|Nodes], LocalTypes, TargetTypes) ->
    RemoteResourceList = call_resource_up(Node, LocalTypes, TargetTypes),
    NewResources       = handle_new_resources(RemoteResourceList, Resources, CallBack),
    sync_inform_nodes(NewResources, CallBack, Nodes, LocalTypes, TargetTypes); 
sync_inform_nodes(Resources, CallBack, [], LocalTypes, TargetTypes) ->
    Resources.

% Call the callback and store resource for each of the new resources returned by call_resource_up/3 
handle_new_resources([{RemoteType, RemoteResource}|T], Resources, CallBack) ->
    do_callback(CallBack, RemoteType, RemoteResource),
    handle_new_resources(T, rd_store:store_resource(RemoteType, RemoteResource, Resources), CallBack);
handle_new_resources([], Resources, CallBack) ->
    Resources.
    

%% Determines if "we" are a module that the client is interested in.
%% Expects: LocalType  - The type of this application.
%%          TargetType - The types that the remote module is interested in.
%%                       This can be a term() | list() of term() | all
%%                       in the case of the atom() all any node running the 
%%                       resource management process will respond.
%% Returns: bool().
can_serve(Type, all) ->
    true;
can_serve(LocalType, []) ->
    false;
can_serve(Type, Type) ->
    true;
can_serve(LocalType, TargetTypes) ->
    lists:any(fun(Type) -> if Type == LocalType -> true; true -> false end end, TargetTypes).
				   
%% Determines if a new resource is one of our target types 
%% Expects: TargetTypes - The types that the er are interested in.
%%                        This can be a term() | list() of term() | all
%%                        in the case of the atom() all any node running the 
%%                        resource management process will respond.
%%          RemoteType  - The type of the remote application.
%% Returns: bool().
can_use(all, _) ->
    true;
can_use([], _) ->
    false;
can_use(Type, Type) ->
    true;
can_use(TargetTypes, RemoteType) ->
    lists:any(fun(Type) -> if Type == RemoteType -> true; true -> false end end, TargetTypes).
				   
%% format_localtypes/1 
%% Returns:
%%  {ok, list() of FullType} | {error, localtypes}.
%%  FullType = {atom(), Resource}.
%%  Resource = most commonly; pid() | node() | {atom(), node()} | GlobalName.
format_localtypes(RawLocalTypes) -> format_localtypes(RawLocalTypes, []).

format_localtypes([{TypeName, Resource}|T], Acc)  when atom(TypeName) -> 
    format_localtypes(T, [{TypeName, Resource}|Acc]);
format_localtypes([TypeName|T], Acc) when atom(TypeName) -> 
    format_localtypes(T, [{TypeName, node()}|Acc]);
format_localtypes({TypeName, Resource}, _) when atom(TypeName) -> 
    {ok, [{TypeName, Resource}]};
format_localtypes(TypeName, _) when atom(TypeName) -> 
    {ok, [{TypeName, node()}]};
format_localtypes([], Acc) -> 
    {ok, Acc};
format_localtypes(_, _) -> 
    {error, localtypes}.


%% Basic formatting for the target types.
%% Returns: list() of TargetTypes
format_targettypes(TargetTypes) when list(TargetTypes) ->
    {ok, TargetTypes};
format_targettypes(TargetTypes) when atom(TargetTypes) ->
    {ok, [TargetTypes]}.



