%%%-------------------------------------------------------------------
%%% File    : rd_heartbeat.erl
%%% Author  : Martin J. Logan 
%%% @doc This does an inform nodes at a specified time interval.
%%% @end
%%%-------------------------------------------------------------------
-module(rd_heartbeat).

-behaviour(gen_server).
%%--------------------------------------------------------------------
%% Include files
%%--------------------------------------------------------------------

%%--------------------------------------------------------------------
%% External exports
-export([start_link/1]).

%% gen_server callbacks
-export([init/1, handle_call/3, handle_cast/2, handle_info/2, terminate/2, code_change/3]).

-define(SERVER, ?MODULE).

-record(state, {frequency, node_contact}).

%%====================================================================
%% External functions
%%====================================================================
%%--------------------------------------------------------------------
%% @doc Starts the server
%% <pre> 
%% Expects:
%%  Frequency - The frequency of heartbeats in milliseconds.
%% </pre>
%% @spec start_link(Frequency) -> {ok, Pid}
%% @end
%%--------------------------------------------------------------------
start_link(Frequency) ->
    gen_server:start_link({local, ?SERVER}, ?MODULE, [Frequency], []).

%%====================================================================
%% Server functions
%%====================================================================

%%--------------------------------------------------------------------
%% Function: init/1
%% Description: Initiates the server
%% Returns: {ok, State}          |
%%          {ok, State, Timeout} |
%%          ignore               |
%%          {stop, Reason}
%%--------------------------------------------------------------------
init([Frequency]) ->
    error_logger:info_msg("heartbeat started: ~p~n", [Frequency]),
    ContactNode = case lists:keysearch(contact_node, 1, init:get_arguments()) of
                      {value, {contact_node, [I_ContactNode]}} -> {ok, list_to_atom(I_ContactNode)};
                      _                                        -> application:get_env(contact_node)
                  end,

    case ContactNode of
        {ok, Node} -> 
            {ok, #state{frequency = Frequency, node_contact = Node}, Frequency};
        undefined  -> 
            {ok, #state{frequency = Frequency, node_contact = null}, Frequency}
    end.

%%--------------------------------------------------------------------
%% Function: handle_call/3
%% Description: Handling call messages
%% Returns: {reply, Reply, State}          |
%%          {reply, Reply, State, Timeout} |
%%          {noreply, State}               |
%%          {noreply, State, Timeout}      |
%%          {stop, Reason, Reply, State}   | (terminate/2 is called)
%%          {stop, Reason, State}            (terminate/2 is called)
%%--------------------------------------------------------------------
handle_call(Request, From, State) ->
    Reply = ok,
    {reply, Reply, State}.

%%--------------------------------------------------------------------
%% Function: handle_cast/2
%% Description: Handling cast messages
%% Returns: {noreply, State}          |
%%          {noreply, State, Timeout} |
%%          {stop, Reason, State}            (terminate/2 is called)
%%--------------------------------------------------------------------
handle_cast(Msg, State) ->
    {noreply, State}.

%%--------------------------------------------------------------------
%% Function: handle_info/2
%% Description: Handling all non call/cast messages
%% Returns: {noreply, State}          |
%%          {noreply, State, Timeout} |
%%          {stop, Reason, State}            (terminate/2 is called)
%%--------------------------------------------------------------------
handle_info(timeout, State) when State#state.node_contact =:= null ->
    resource_discovery:inform_network(),
    {noreply, State, State#state.frequency};
handle_info(timeout, State) ->
    error_logger:info_msg("heartbeating : ~p~n", [State#state.frequency]),
    net_adm:ping(State#state.node_contact),
    resource_discovery:inform_network(),
    {noreply, State, State#state.frequency}.

%%--------------------------------------------------------------------
%% Function: terminate/2
%% Description: Shutdown the server
%% Returns: any (ignored by gen_server)
%%--------------------------------------------------------------------
terminate(Reason, State) ->
    ok.

%%--------------------------------------------------------------------
%% Func: code_change/3
%% Purpose: Convert process state when code is changed
%% Returns: {ok, NewState}
%%--------------------------------------------------------------------
code_change(OldVsn, State, Extra) ->
    {ok, State}.

%%--------------------------------------------------------------------
%%% Internal functions
%%--------------------------------------------------------------------
