%%%-------------------------------------------------------------------
%%% File    : rd_store.erl
%%% Author  : Martin J. Logan <martin@gdubya.botomayo>
%%% @doc The storage management functions for resource discovery
%%% @end
%%%-------------------------------------------------------------------
-module(rd_store).

%%--------------------------------------------------------------------
%% Include files
%%--------------------------------------------------------------------

%%--------------------------------------------------------------------
%% External exports
%%--------------------------------------------------------------------

% Create
-export([
         new_LPS/0,
         new_RL/0
        ]).

% Lookup
-export([
         round_robin_lookup/2,
         index_lookup/3,
         lookup_localtypes/0,
         lookup_targettypes/0,
         lookup_all_resources/2,
         lookup_num_resources/2,
         lookup_num_types/1,
         lookup_types/1
        ]).

% Delete
-export([
         delete_localtype/1,
         delete_targettype/1,
         delete_resource/3
        ]).

% Store
-export([
         store_localtype/1,
         store_localtypes/1,
         store_targettype/1,
         store_targettypes/1,
         store_resource/3
        ]).

%%--------------------------------------------------------------------
%% Macros
%%--------------------------------------------------------------------
-define(LT, rd_local_types).
-define(TT, rd_target_types).

%%====================================================================
%% External functions
%%====================================================================

%%%---------------------------
%%% Local Type, Target type Storage
%%%---------------------------

%%-----------------------------------------------------------------------
%% @doc LPS stands for "Local Parameter Store" 
%% Initialises persistent storage.
%% @end
%%-----------------------------------------------------------------------
new_LPS() ->
    ets:new(?LT, [named_table, public]),
    ets:new(?TT, [named_table, public]).

%%-----------------------------------------------------------------------
%% @doc Store the "I haves" or localtypes for resource discovery.
%% <pre>
%% Expects:
%%  LocalFullTypeList - list() of FullTypes
%%   FullTypes = {atom(), Resource}
%%   Resource = most commonly; pid() | node() | {atom(), node()} | GlobalName.
%% </pre>
%% @spec store_localtypes(LocalFullTypeList) -> ok
%% @end
%%-----------------------------------------------------------------------
store_localtypes(LocalFullTypeList) ->
    lists:foreach(fun(FullType) -> store_localtype(FullType) end, LocalFullTypeList).
     
%%-----------------------------------------------------------------------
%% @doc Stores a value in the persistent storage.
%% <pre>
%% Expects: 
%%  FullType - {LocalType, Resource}
%%  LocalType = atom()
%%  Resource  = term()
%% </pre>
%% @spec store_localtype(FullType) -> ok
%% @end
%%-----------------------------------------------------------------------
store_localtype({LocalType, Resource}) ->
    ets:insert(?LT, {LocalType, Resource}).

%%-----------------------------------------------------------------------
%% @doc Output the localtypes.
%% @spec lookup_localtypes() -> LocalTypeList
%% @end
%%-----------------------------------------------------------------------
lookup_localtypes() ->
    ets:match_object(?LT, '$1').

%%-----------------------------------------------------------------------
%% @doc Remove a localtype.
%% <pre>
%% Expects:
%%  LocalType = atom()
%% </pre>
%% @spec delete_localtype(LocalType) -> true
%% @end
%%-----------------------------------------------------------------------
delete_localtype(LocalType) ->
    ets:delete(?LT, LocalType).


%%-----------------------------------------------------------------------
%% @doc Store one or many target type or "I want" values.
%% <pre>
%% Expects:
%%  LocalFullTypeList - list() of FullTypes
%%   FullTypes = {atom(), Resource}
%%   Resource = most commonly; pid() | node() | {atom(), node()} | GlobalName.
%% </pre>
%% @spec store_targettypes(TargetTypeList) -> ok
%% @end
%%-----------------------------------------------------------------------
store_targettypes(TargetTypeList) ->
    lists:foreach(fun(TargetType) -> store_targettype(TargetType) end, TargetTypeList).
     
%%-----------------------------------------------------------------------
%% @doc Stores a target type value in the persistent storage.
%% <pre>
%% Expects:
%%  LocalType - atom()
%%  Resource  - term()
%% </pre>
%% @spec store_targettype(TargetType) -> ok
%% @end
%%-----------------------------------------------------------------------
%% XXX Ugly need elegant solution
store_targettype(TargetType) ->
    ets:insert(?TT, {TargetType}).

%%-----------------------------------------------------------------------
%% @doc Output the localtypes.
%% @spec lookup_targettypes() -> TargetTypeList
%% @end
%%-----------------------------------------------------------------------
lookup_targettypes() ->
    lists:foldl(fun({TargetType}, Acc) -> [TargetType|Acc] end, [], ets:match_object(?TT, '$1')).

%%-----------------------------------------------------------------------
%% @doc Deletes a target type.
%% <pre>
%% Expects:
%%  TargetType = atom()
%% </pre>
%% @spec delete_targettype(TargetType) -> true
%% @end
%%-----------------------------------------------------------------------
delete_targettype(TargetType) ->
    ets:delete(?TT, TargetType).

%%%---------------------------
%%% Network Resource Storage
%%%---------------------------

%%-----------------------------------------------------------------------
%% @doc Returns an empty store for resources.
%% @end
%%-----------------------------------------------------------------------
new_RL() ->
    dict:new().

%%-----------------------------------------------------------------------
%% @doc Gets resource of a particular type at a given index.
%% <pre>
%% Expects:
%%  Type   - The type of resource you want.
%%  Struct - The resource structure.
%% </pre>
%% @spec index_lookup(Type, Index, Struct) -> {error, Struct} | {ok, Resource}
%% @end
%%-----------------------------------------------------------------------
index_lookup(Type, Index, Struct) ->
    case dict:find(Type, Struct) of
	{ok, RL} -> get_nth(Index, RL, Struct);
	{ok, []} -> {error, Struct};
	error    -> {error, Struct}
    end.

get_nth(N, List, Struct) ->
    if 
        length(List) >= N -> 
            {ok, lists:nth(N, List)};
        true -> 
            {error, Struct}
    end.

%%-----------------------------------------------------------------------
%% @doc Gets resource of a particular type outputs and places it in last position.
%% <pre>
%% Expects:
%%  Type   - The type of resource you want.
%%  Struct - The resource structure.
%% </pre>
%% @spec round_robin_lookup(Type, Struct) -> 
%% 	{error, Struct} | {ok, {Resource, NewStruct}}
%% @end
%%-----------------------------------------------------------------------
round_robin_lookup(Type, Struct) ->
    case dict:find(Type, Struct) of
	{ok, [Resource|RL]} -> {ok, {Resource, dict:store(Type, (RL ++ [Resource]), Struct)}};
	{ok, []}            -> {error, Struct};
	error               -> {error, Struct}
    end.
			       
%%-----------------------------------------------------------------------
%% @doc Adds a new type and value.
%% <pre>
%% Variables:
%%  Type   - The type of resource you want.
%%  Value  - The value to be removed
%%  Struct - The resource structure.
%% </pre>
%% @spec store_resource(Type, Resource, Struct) -> NewStruct
%% @end
%%-----------------------------------------------------------------------
store_resource(Type, Resource, Struct) ->
    case dict:find(Type, Struct) of
	{ok, []} -> 
            dict:store(Type, [Resource], Struct);
	{ok, RL} -> 
            case lists:member(Resource, RL) of
                true  -> Struct;
                false -> dict:store(Type, RL ++ [Resource], Struct)
            end;
	error -> 
            dict:store(Type, [Resource], Struct)
    end.



%%-----------------------------------------------------------------------
%% @doc Remove a single value from the resource list.
%% <pre>
%% Variables:
%%  Type   - The type of resource you want.
%%  Value  - The value to be removed
%%  Struct - The resource structure.
%% </pre>
%% @spec delete_resource(Type, Value, Struct) -> NewStruct
%% @end
%%-----------------------------------------------------------------------
delete_resource(Type, Value, Struct) ->
    case dict:find(Type, Struct) of
	{ok, []} -> Struct;
	{ok, RL} -> dict:store(Type, lists:delete(Value, RL), Struct);
	error    -> Struct
    end.
    
%%-----------------------------------------------------------------------
%% @doc Outputs a list of all resources for a particular type.
%% <pre>
%% Expects:
%%  Type   - The type of resource you want.
%%  Struct - The resource structure.
%% </pre>
%% @spec lookup_all_resources(Type, Struct) -> ResourceList
%% @end
%%-----------------------------------------------------------------------
lookup_all_resources(Type, Struct) ->
    case dict:find(Type, Struct) of
	{ok, RL} -> RL;
	error    -> []
    end.

%%-----------------------------------------------------------------------
%% @doc Outputs the number of resource types.
%% <pre>
%% Expects:
%%  Struct = The resource structure.
%% </pre>
%% @spec lookup_num_types(Struct) -> integer()
%% @end
%%-----------------------------------------------------------------------
lookup_num_types(Struct) ->
    length(dict:fetch_keys(Struct)).

%%-----------------------------------------------------------------------
%% @doc Outputs the types of resources.
%% <pre>
%% Expects: 
%%  Struct = The resource structure.
%% </pre>
%% @spec lookup_types(Struct) -> [Types]
%% @end
%%-----------------------------------------------------------------------
lookup_types(Struct) ->
    dict:fetch_keys(Struct).

%%-----------------------------------------------------------------------
%% @doc Outputs the number of resources of a particualr type.
%% <pre>
%% Expects: Struct
%%  Struct = The resource structure.
%% </pre>
%% @spec lookup_num_resources(Type, Struct) -> integer()
%% @end 
%%-----------------------------------------------------------------------
lookup_num_resources(Type, Struct) ->
    case dict:find(Type, Struct) of
	{ok, []} -> 0;
	{ok, RL} -> length(RL);
	_        -> 0
    end.
