%%%-------------------------------------------------------------------
%%% File    : resource_discovery.erl
%%% Author  : Martin J. Logan <martin@gdubya.botomayo>
%%% @doc 
%%%
%%% <h2>Description</h2>
%%% <p>
%%% The Resource Discovery application centers around the notion of a resource(). A resource is the coupling of a type() and an instance() of that type().
%%% For example: {parser, parser_node@mecha.erlware.com} would be a resource() with type parser and instance parser_node@mecha... 
%%% Resource Discovery allows a client to find resources in an erlang cluster of nodes. A client will inform the system of two things; "local resources"
%%% and "target types". These are basically the "I haves" and "I wants" in otherwords local resources are the resource()s that are to be found on the 
%%% local node and the target types are the types of resources that the client would like to discover across all nodes in the cluster. Resource Discovery is
%%% a simple application and intended to stay that way. Most of the tracking and managing of resources is left up to developers to implement in their own 
%%% applications. Resource Discovery does have some facility for managing common_resources() see the section entitled "Managed Common Resources"
%%% below for more details. 
%%% </p> 
%%% 
%%% <p>To start <em>Resource Discovery</em> it must be included in the
%%% release for the particular app or apps that wish to use it. </p>
%%% 
%%% <p><strong>Contact Node</strong> A contact node is the default way to bring a node into a cluster of nodes. Resource Discovery 
%%% can be configured via otp enviornment vars to ping a given node on startup. This is useful for bringing a new app into an otp 
%%% cluster. the line in configuration is: 
%%% <code>
%%%   {contact_node, 'Name@Node'} 
%%% </code>
%%% A contact node can also be specified by using the command line flag:
%%% <code>
%%%   -contact_node Name@Node 
%%% </code>
%%% If a <code>contact_node</code> is supplied and cannot be reached then Resource
%%% Discovery will die.</p>
%%% 
%%% <strong>The "here is what I have" heartbeat</strong><br/>
%%% <p>Another configuration option prompts the <em>resource_discovery</em>
%%% to periodically call the <code>inform_nodes/0</code> api function. The period is specified by the
%%% integer in the value tuple. The units are miliseconds. If a
%%% <code>contact_node</code> has been specified then it will also ping that
%%% node at the same frequency.
%%% <code>
%%%   {heartbeat, {true, 60000}},
%%% </code></p>
%%% 
%%% <p>It is also possible to use an enviornment variable to specify a
%%% callback module. This module will export the function
%%% <code>resource_up</code> with arity two. This function will be called
%%% every time a target resource makes itself known. This includes
%%% re-anouncements as a result of turning on the heartbeat mechanism.
%%% See <code>resource_up/2</code> in the <em>Callback Functions </em> 
%%% section of this doc.</p>
%%
%%% <h2>Callback Functions</h2>
%%%
%%% <code>resource_up(type(), instance()) -> void()</code>
%%%
%%% <p> This callback function is called every time a resource that is a member of
%%% <code>TargetTypes</code> is brought to the attention of the 
%%% resource discovery. This is true of initial network presence notifications
%%% as well as notifications due to hearbeats.</p>
%%% 
%%% <h2>Configuration Specs</h2>
%%%
%%% <p> The config options for resource_discovery.</p>
%%%
%%% Heartbeat:
%%% <code> 
%%%  {heartbeat, {bool(), MiliSeconds}}
%%% </code><br/>
%%%
%%% Initial Contact:
%%% <code>
%%%   {contact_node, NodeName}
%%% </code><br/>
%%%
%%% Callback Module:
%%% <code>
%%%   {callback, atom()}
%%% </code>
%%%
%%% <h2>Managed Common Resources</h2>
%%% <p>
%%% This functionality manages common_resource()s. All things are done as normally with resource discovery exept that 
%%% target types are added through add_managed_targettypes API function. The instance()s associated with these type()s must
%%% be common_instance()s. This application will then take over monitoring these instance()s as it becomes aware of them through resource_discovery.
%%% When a resource is no longer available as is determined by a {nodedown, node()} message from erlang:monitor_node/2 or a 'DOWN' message
%%% from erlang:monitor/2 that common_resource() is automatically deleted from the resource_discovery cache of resources. 
%%% </p> 
%%% 
%%% <pre>
%%% Explanation of Appwide Types:
%%%  resource()        - A common resource is a resource of any type() coupled with an instance()
%%%  common_resource() - A common resource is a resource of any type() but with an instance()=common_instance() i.e an instance() 
%%%                      that is either a node() or a pid()
%%%  type()            - A name for a group of resources. mnesia_node would be a valid type to identify all instances of mnesia dbs on the network.
%%%  instance()        - The specific term that a type refers to. Many instance()s can belong under a type() grouping.
%%%  common_instance() - The specific instance of a particular type. This common_instance() must be either a node() or a pid(). 
%%%                      for example; mnesia_one@mecha.erlware.com might be one instance of type()=mnesia_node
%%%
%%% Formal Definition of Appwide Types:
%%%  resource()          = common_resource() | {type(), instance()} | type()
%%%   common_resource()  = {type(), common_instance()} | type()
%%%   type()             = atom() 
%%%   instance()         = common_instance() | term()
%%%    common_instance() = node() | pid()
%%% </pre>
%%% @end
%%%-------------------------------------------------------------------
-module(resource_discovery).

%%--------------------------------------------------------------------
%% External exports
%%--------------------------------------------------------------------

% Standard exports.
-export([
         start/2,
    	 start_link/0, 
    	 start_link/1, 
         stop/1, 
         stop/0 
        ]).

% Add
-export([
         add_localtypes/1,
         add_targettypes/1,
         add_managed_targettypes/1,
         add_callback/1
        ]).

% Get
-export([
         get_resource/1, 
         get_resource/2, 
         get_all_resources/1, 
         get_num_resource/1, 
         get_types/0,
         get_num_types/0,
         get_contact_node/0
        ]).

% Remove
-export([
         remove_localtype/1,
         remove_targettype/1,
         remove_resource/2 
        ]).

% Other
-export([
         full_add/2,
         inform_network/0,
         sync_inform_network/0,
         contact_node/0,
         start_heartbeat/0,
         start_heartbeat/1,
         stop_heartbeat/0,
         execute_with_resource/2,
         execute_with_same_resource/2,
         rpc_call_with_discovery/4
        ]).

%%--------------------------------------------------------------------
%% Macros
%%--------------------------------------------------------------------
-define(RD, ?MODULE).

%%====================================================================
%% External functions
%%====================================================================

%%--------------------------------------------------------------------
%% @doc Starts the resource discovery application.
%% @spec start(Type, StartArgs) -> {ok, Pid} | {ok, Pid, State} | {error, Reason}
%% @end
%%--------------------------------------------------------------------
start(Type, StartArgs) ->
    case rd_sup:start_link(StartArgs) of
        {ok, Pid} ->
            {ok, Pid};
        Error ->
            Error
    end.

%%------------------------------------------------------------------------------
%% @doc Starts the resource discovery processes.
%% <pre>
%%
%% Expects: 
%%  Options = [Options]
%%   Options = {heartbeat, MiliSeconds}  
%%
%% </pre>
%%
%% @spec start_link(Options) -> {ok, pid()} | {error, Reason}
%% @end
%%------------------------------------------------------------------------------
start_link(Options) ->
    gen_server:start_link({local, ?RD}, rd_core, [Options], []).

%% @spec start_link() -> {ok, pid()} | {error, Reason}
%% @equiv start_link(Options)
start_link() ->
    gen_server:start_link({local, ?RD}, rd_core, [], []).

%%------------------------------------------------------------------------------
%% @doc Adds to the list of managed target types. Specifying that a target type() is managed through this api causes any resource()s 
%% of that type() to be monitored by Resource Discovery until they "die". When Resource Discovery notices that a managed resource instance() 
%% had died it is automatically removed from the cache of available resources.  For a resource to be managed it must be either a node() or a pid().
%% <pre>
%%
%% Expects: 
%%  TargetTypes  - A type or list of type()s to be discovered in the node cluster
%%
%% Types:
%%  TargetTypes = [type()] | type()
%%  Reason = term()
%%
%% </pre>
%%
%% @spec add_managed_targettypes(TargetTypes) -> ok | {error, Reason}
%% @end
%%------------------------------------------------------------------------------
add_managed_targettypes(TargetTypes) -> 
    rd_common_resource_manager:add_targettypes(TargetTypes).
    
%%------------------------------------------------------------------------------
%% @doc Adds to the list of target types.
%% <pre>
%%
%% Expects: 
%%  TargetTypes  - A type or list of type()s to be discovered in the node cluster
%%
%% Types:
%%  TargetTypes = [type()] | type()
%%  Reason = term()
%%
%% </pre>
%%
%% @spec add_targettypes(TargetTypes) -> ok | {error, Reason}
%% @end
%%------------------------------------------------------------------------------
add_targettypes(TargetTypes) -> 
    gen_server:call(?RD, {add_targettypes, TargetTypes}).

%%------------------------------------------------------------------------------
%% @doc Adds to the list of local resources.
%% <pre>
%%
%% Expects: 
%%  LocalResources  - A list of local resources.
%%  
%% Types:
%%  LocalResources = [type() | resource()] | resource() | type() 
%%  Reason = term()
%%
%%</pre>
%%
%% @spec add_localtypes(LocalResources) -> ok | {error, Reason}  
%% @end
%%------------------------------------------------------------------------------
add_localtypes(LocalResources) -> 
    gen_server:call(?RD, {add_localtypes, LocalResources}).

%%------------------------------------------------------------------------------
%% @doc Add a callback to the list of callbacks to be called upon new resources entering the system.
%% <pre>
%% Variables:
%%  Module - The module that the callback resource_up/2 is exported from.
%%
%% Types:
%%  Module = atom()
%% </pre>
%% @spec add_callback(Module) -> ok | exit()
%% @end
%%------------------------------------------------------------------------------
add_callback(Module) ->
    gen_server:call(?RD, {add_callback, Module}).

%%------------------------------------------------------------------------------
%% @doc Stops the server with the reason provided.
%% <pre>
%%
%% Types:
%%  Reason = atom() | string()
%%
%% </pre>
%%
%% @spec stop(Reason) -> ok
%% @end
%%------------------------------------------------------------------------------
stop(Reason) ->
    gen_server:cast(?RD, {stop, Reason}).

%%------------------------------------------------------------------------------
%% @doc Stops the server with the reason 'normal'.
%% @spec stop() -> ok
%% @end
%%------------------------------------------------------------------------------
stop() ->
    gen_server:cast(?RD, {stop, normal}).

%%------------------------------------------------------------------------------
%% @doc Returns a resource.
%% <pre>
%% 
%% Note: This function round robins the resources of a particular type.
%%
%% Types: 
%%  Type = atom()
%%
%% </pre>
%%
%% @spec get_resource(Type) -> {ok, Resource} | {error, no_resources}
%% @end
%%------------------------------------------------------------------------------
get_resource(Type) ->
    gen_server:call(?RD, {get_resource, Type}). 

%%------------------------------------------------------------------------------
%% @doc Replies with the resource at the index specified. If for example we had a list of resources that looked like {R1, R2, R3} and index 2 was requested 
%% the user would receive R2.
%% <pre>
%%
%% Note: This function does not re-order the resources of  the type specified
%% 
%% Types:
%%  Index = integer()
%%
%% </pre>
%%
%% @spec get_resource(Type, Index) -> {ok, Resource} | {error, no_resources}
%% @end
%%------------------------------------------------------------------------------
get_resource(Type, Index) ->
    gen_server:call(?RD, {get_resource, Type, Index}). 

%%------------------------------------------------------------------------------
%% @doc Returns ALL resources for a particular type.
%% <pre>
%%
%% Types:
%%  ResourceList = [resource()]
%%
%% </pre>
%% @spec get_all_resources(type()) -> {ok, ResourceList}
%% @end
%%------------------------------------------------------------------------------
get_all_resources(Type) ->
    gen_server:call(?RD, {get_all_resources, Type}). 

%%------------------------------------------------------------------------------
%% @doc Removes a resource from the resource pool.
%% @spec remove_resource(type(), resource()) -> ok
%% @end
%%------------------------------------------------------------------------------
remove_resource(Type, Resource) ->
    gen_server:cast(?RD, {remove_resource, Type, Resource}).

%%------------------------------------------------------------------------------
%% @doc Counts a particular resource type.
%% @spec get_num_resource(type()) -> {ok, integer()}
%% @end
%%------------------------------------------------------------------------------
get_num_resource(Type) ->
    gen_server:call(?RD, {get_num_resource, Type}). 

%%------------------------------------------------------------------------------
%% @doc Remove a target type.
%% @spec remove_targettype(type()) -> {ok, integer()}
%% @end
%%------------------------------------------------------------------------------
remove_targettype(Type) ->
    gen_server:cast(?RD, {remove_targettype, Type}),
    rd_common_resource_manager:remove_targettype(Type).

%%------------------------------------------------------------------------------
%% @doc Remove a local resource.
%% <pre>
%%
%% Expects:
%%  LocalResource - Either a type or a full resource to be removed. If only the type() is provided the instance() is assumed to be the local node.
%%
%% Types:
%%  LocalResource = type() | resource()
%%
%% </pre>
%%
%% @spec remove_localtype(LocalResource) -> void()
%% @end
%%------------------------------------------------------------------------------
remove_localtype(LocalResource) ->
    gen_server:call(?RD, {remove_localtype, LocalResource}). 

%%------------------------------------------------------------------------------
%% @doc Gets a list of the types that are being managed.
%% @spec get_types() -> {ok, list()}
%% @end
%%------------------------------------------------------------------------------
get_types() ->
    gen_server:call(?RD, get_types). 

%%------------------------------------------------------------------------------
%% @doc Gets the number of resource types locally known.
%% @spec get_num_types() -> {ok, integer()}
%% @end
%%------------------------------------------------------------------------------
get_num_types() ->
    gen_server:call(?RD, get_num_types). 

%%------------------------------------------------------------------------------
%% @doc inform the network of your existance and all the resources that are provided by the local node as well as all the target types that the local
%% instance of Resource Discovery would like to know about.
%% *NOTE* When pinging into a cluster the full connection of all nodes in the cluster is not synchronous to the net_adm:ping call. It is wise
%%        to wait a few seconds inbetween pinging into a cluster and informing the network so as to ensure that the cloud is fully connected. 
%% @spec inform_network() -> ok | exit(Reason)
%% @end
%%------------------------------------------------------------------------------
inform_network() ->
    gen_server:call(?RD, inform_network). 

%%------------------------------------------------------------------------------
%% @doc Inform the network of your existance and wait for a response from all remote nodes informed.
%% <p>When this call returns all available remote, full targettypes are guaranteed to be locally cached.
%% Calling this function ensures the pruning of all dead resources, the cache of resources is totally up 
%% date after calling this function.</p>
%% *NOTE* When pinging into a cluster the full connection of all nodes in the cluster is not synchronous to the net_adm:ping call. It is wise
%%        to wait a few seconds inbetween pinging into a cluster and informing the network so as to ensure that the cloud is fully connected. 
%% @spec sync_inform_network() -> ok | exit(Reason)
%% @end
%%------------------------------------------------------------------------------
sync_inform_network() ->
    gen_server:call(?RD, sync_inform_network). 

%%------------------------------------------------------------------------------
%% @doc Add a new list of localtypes target types and inform the network.
%% <pre>
%%
%% Expects: 
%%  LocalResource - Either a type or a full resource to be removed. If only the type() is provided the instance() is assumed to be the local node.
%%  TargetTypes  - A type or list of types to be discovered in the node cluster
%%
%% Types:
%%  LocalResources = [type() | resource()] | resource() | type() 
%%  TargetTypes = [type()] | type()
%%
%% </pre>
%%
%% @spec full_add(LocalResources, TargetTypes) -> ok | exit(Reason)
%% @end
%%------------------------------------------------------------------------------
full_add(LocalResources, TargetTypes) ->
    gen_server:call(?RD, {full_add, LocalResources, TargetTypes}). 

%%------------------------------------------------------------------------------
%% @doc Starts the resource discovery heartbeat process.
%% <pre>
%%
%% Causes the resource_discovery to perform a periodic 
%% resource_discovery:inform_network
%%
%% Note: This can also be a configuration option. 
%% <code>
%%    {heartbeat, {bool(), MiliSeconds}},
%% </code>
%%
%% Note: If no time is specified the time defaults to 60000 ms, or one minute.
%%
%% Types: 
%%  MiliSeconds = integer()
%%
%% </pre>
%%
%% @spec start_heartbeat(MiliSeconds) -> void() 
%% @end
%%------------------------------------------------------------------------------
start_heartbeat(MiliSeconds) ->
    rd_sup:start_heartbeat(MiliSeconds).

%% @spec start_heartbeat() -> void() 
%% @equiv start_heartbeat(MiliSeconds)
start_heartbeat() ->
    start_heartbeat(60000).

%%------------------------------------------------------------------------------
%% @doc Stops the heart beat process.
%% @spec stop_heartbeat() -> void()
%% @end
%%------------------------------------------------------------------------------
stop_heartbeat() ->
    rd_sup:stop_heartbeat().
    
%%------------------------------------------------------------------------------
%% @doc Contacts resource discoveries initial contact node.
%% <pre>
%%
%% The initial contact node is specified in configuration with:
%% <code>
%%   {contact_node, NodeName}
%% </code>
%%
%% Types:
%%  Node = node()
%%  Result = pong | pang.
%%
%% </pre>
%%
%% @spec contact_node() -> {Node, Result}
%% @end
%%------------------------------------------------------------------------------
contact_node() ->
    gen_server:call(?RD, contact_node, 10000).
    
%%------------------------------------------------------------------------------
%% @doc Get the contact node for the application.
%% @spec get_contact_node() -> {ok, Value} | undefined
%% @end
%%------------------------------------------------------------------------------
get_contact_node() ->
    application:get_env(resource_discovery, contact_node).
    
%%------------------------------------------------------------------------------
%% @doc Execute a fun on a resource until it returns true or there are no more resources.
%% If the fun returns false the resource that caused the failure will be removed from the
%% cache of available resources. This function uses the function get_resource/1.
%% This is a convenience function.
%% <pre>
%% Varibles:
%%  Type - The resource type to get from resource discovery.
%%  Fun - The fun to execute. It is of arity one. The fun() must return a bool(). 
%%        Example Fun = fun(Resource) -> ok == distributed_data_store:store(Resource, Data) end.
%% </pre>
%% @spec execute_with_resource(Resource, Fun) -> ok | error
%% @end
%%------------------------------------------------------------------------------
execute_with_resource(Type, Fun) ->
    case get_resource(Type) of
        {ok, Resource} ->
            case Fun(Resource) of
                true -> 
                    ok;
                false ->
                    remove_resource(Type, Resource),
                    execute_with_resource(Type, Fun)
            end;
        {error, no_resources} ->
            error
    end.

%%------------------------------------------------------------------------------
%% @doc Execute a fun on a resource until it returns true or there are no more resources.
%% If the fun returns false the resource that caused the failure will be removed from the
%% cache of available resources. This function uses the function get_resource/2.
%% This is a convenience function.
%% <pre>
%% Varibles:
%%  Type - The resource type to get from resource discovery.
%%  Fun - The fun to execute. It is of arity one. The fun() must return a bool(). 
%%        Example Fun = fun(Resource) -> ok == distributed_data_store:store(Resource, Data) end.
%% </pre>
%% @spec execute_with_same_resource(Resource, Fun) -> ok | error
%% @end
%%------------------------------------------------------------------------------
execute_with_same_resource(Type, Fun) ->
    case get_resource(Type, 1) of
        {ok, Resource} ->
            case Fun(Resource) of
                true -> 
                    ok;
                false ->
                    remove_resource(Type, Resource),
                    execute_with_resource(Type, Fun)
            end;
        {error, no_resources} ->
            error
    end.



%%------------------------------------------------------------------------------
%% @doc Execute an rpc on a resource.  If the result of the rpc is {badrpc, reason} the 
%%      resource is removed and the next resource is tried, else the result is 
%%      returned to the user.
%% <pre>
%% Varibles:
%%  Type - The resource type to get from resource discovery.
%% </pre>
%% @spec rpc_with_discovery(Type, Module, Function, Args) -> RPCResult | {error, no_resources}
%% @end
%%------------------------------------------------------------------------------
rpc_call_with_discovery(Type, Module, Function, Args) ->
    case resource_discovery:get_resource(Type) of
	{ok, Resource} -> 
	    io:format("got a resource ~p~n", [Resource]),
	    case rpc:call(Resource, Module, Function, Args) of
		{badrpc, Reason} ->
		    io:format("got a badrpc ~p~n", [Reason]),
		    resource_discovery:remove_resource(Type, Resource),
		    rpc_call_with_discovery(Type, Module, Function, Args);
		Reply ->
		    io:format("result of rpc was ~p~n", [Reply]),
		    Reply
	    end;
        {error, no_resources} -> 
	    {error, no_resources}
    end.
